from LTSPAgent.plugin import Plugin, auto_update
from storm.locals import *
import datetime, xmlrpclib, paramiko, re, sys, logging, os, codecs
from hashlib import sha1

class Account(object):
    __storm_table__ = "account"
    id = Int(primary=True)
    server = Unicode()
    username = Unicode()
    password = Unicode()
    fullname = Unicode()
    source = Unicode()
    session = Unicode()
    enabled = Bool()
    created = DateTime()

class vmmanager(Plugin):
    unavailable_servers=[]
    locales={}
    locales_description={}
    packages={}

    def init_plugin(self):
        # Disable paramiko logging
        logging.getLogger('paramiko.transport').setLevel(logging.CRITICAL)

        # Initialize the shared DB connection
        if not "postgres://" in self.get_config_path(self.config,"general","database"):
            self.LOGGER.critical("vmmanager only supports PostgreSQL")
            sys.exit(1)

        self.db=create_database(self.get_config_path(self.config,"general","database"))

        # Get the store
        store=Store(self.db)

        # Create the TABLE if it doesn't exist
        try:
            store.execute("CREATE TABLE account (id SERIAL, server VARCHAR, username VARCHAR, password VARCHAR, fullname VARCHAR, source VARCHAR, session VARCHAR, enabled BOOLEAN, created timestamp without time zone)")
            store.commit()
        except:
            pass

        # Extend the schema for TABLE with missing 'source' or 'session' field
        try:
            store.execute("ALTER TABLE account ADD COLUMN source VARCHAR")
            store.execute("ALTER TABLE account ADD COLUMN session VARCHAR")
            store.commit()
        except:
            pass

        # Extend the schema for TABLE with missing 'password' field
        try:
            store.execute("ALTER TABLE account ADD COLUMN password VARCHAR")
            store.commit()
        except:
            pass

        # Load locale nice names
        for path in ("/usr/share/ltsp-agent/plugins/vmmanager/locales.list","plugins/vmmanager/locales.list"):
            if os.path.exists(path):
                for line in codecs.open(path,'r','utf-8').readlines():
                    line=line.strip()
                    self.locales_description[line.split()[0]]=" ".join(line.split()[1:])
                break

        # Call parent function (start the threads)
        Plugin.init_plugin(self)

    def get_ssh(self,serverid):
        # Connect to server
        ssh = paramiko.SSHClient()
        ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        ssh.connect(
            self.get_config_path(self.config,"servers",serverid,"server"),
            port=int(self.get_config_path(self.config,"servers",serverid,"port")),
            username="root",
            password=self.get_config_path(self.config,"servers",serverid,"password"),
            allow_agent=False,
            look_for_keys=False
        )
        return ssh

    def create_user(self,serverid,username,fullname,password,source,session):
        # Get the store
        store=Store(self.db)

        if serverid not in self.get_config_path(self.config,"servers") or serverid in self.unavailable_servers:
            # Invalid server
            return 7

        if not re.match("^[A-Za-z0-9 ]*$",fullname):
            # Invalid fullname, must only contain alphanumeric characters and spaces
            return 3

        if not re.match("^[a-z]*$",username) or username in self.get_config_path(self.config,"general","blacklist"):
            # Invalid login, must only contain lowercase letters
            return 4

        if not re.match("^[A-Za-z0-9]*$",password):
            # Invalid password, must contain only alphanumeric characters
            return 5

        if store.find(Account, server=unicode(serverid), enabled=True).count() >= int(self.get_config_path(self.config,"servers",serverid,"userlimit")):
            # Reached user limit, return false
            return 1

        if store.find(Account, server=unicode(serverid), enabled=True, username=unicode(username)).count() == 1:
            user = store.find(Account, server=unicode(serverid), enabled=True, username=unicode(username))[0]

            if user.password == sha1(password).hexdigest():
                # Account already exists, renew it
                user.created = datetime.datetime.now()
                store.commit()
                return [self.get_config_path(self.config,"servers",serverid,"server"),int(self.get_config_path(self.config,"servers",serverid,"port"))]
            else:
                # Different user already exists
                return 2

        # Add user to the database
        user=Account()
        user.username = unicode(username)
        user.fullname = unicode(fullname)
        user.server = unicode(serverid)
        user.source = unicode(source)
        user.session = unicode(session)
        user.password = unicode(sha1(password).hexdigest())
        user.enabled = True
        user.created = datetime.datetime.now()

        try:
            # Connect to server
            ssh = self.get_ssh(serverid)

            # Create the account
            stdin, stdout, stderr = ssh.exec_command("adduser --quiet --gecos '%s' %s" % (fullname,username))
            stdin.write('%s\n' % password)
            stdin.flush()
            stdin.write('%s\n' % password)
            stdin.flush()
            stdout.read()

            # Disconnect from server
            ssh.close()

            # Save the new user
            store.add(user)
            store.commit()

            # Everything worked, return true
            return [self.get_config_path(self.config,"servers",serverid,"server"),int(self.get_config_path(self.config,"servers",serverid,"port"))]

        except:
            self.LOGGER.error("Server '%s' is unavailale." % serverid)
            self.unavailable_servers.append(serverid)
            return False

    def delete_user(self,serverid,username):
        # Get the store
        store=Store(self.db)

        if serverid not in self.get_config_path(self.config,"servers") or serverid in self.unavailable_servers:
            # Invalid server
            return False

        user=store.find(Account, server=unicode(serverid), enabled=True, username=unicode(username))

        if user.count() != 1:
            return False

        try:
            # Connect to server
            ssh = self.get_ssh(serverid)

            # Disconnect and delete the account
            ssh.exec_command("nxserver --force-terminate %s" % username)
            ssh.exec_command("pkill -9 -u %s" % username)
            ssh.exec_command("umount /home/%s/.gvfs" % username)
            ssh.exec_command("find /dev/shm /var /tmp -user \"%s\" | xargs -r rm -Rf" % username)
            ssh.exec_command("deluser --force --remove-home --quiet %s" % username)

            # Disconnect from server
            ssh.close()

            # Disable account in database
            user[0].enabled = False
            store.commit()

            # Everything worked, return true
            return True

        except:
            self.LOGGER.error("Server '%s' is unavailale." % serverid)
            self.unavailable_servers.append(serverid)
            return False

    @auto_update(3600)
    def update_locales(self):
        for serverid in self.get_config_path(self.config,"servers"):
            if serverid in self.unavailable_servers:
                continue

            try:
                # Connect to the server
                ssh = self.get_ssh(serverid)

                # Get supported locales
                locales=[]
                for line in ssh.exec_command("locale -a")[1].readlines():
                    line=line.strip()
                    if line.endswith('utf8'):
                        if line.split('.')[0] in self.locales_description:
                            nice=self.locales_description[line.split('.')[0]]
                        else:
                            nice=line
                        locales.append((line,nice))

                # Disconnect from server
                ssh.close()

                if len(locales) > 0:
                    self.locales[serverid]=locales
            except:
                self.LOGGER.error("Server '%s' is unavailale." % serverid)
                self.unavailable_servers.append(serverid)

    @auto_update(3600)
    def update_packages(self):
        for serverid in self.get_config_path(self.config,"servers"):
            if serverid in self.unavailable_servers:
                continue

            try:
                # Connect to the server
                ssh = self.get_ssh(serverid)

                # Get package list
                packages=[]
                for line in ssh.exec_command("dpkg -l")[1].readlines():
                    line=line.strip()
                    if line.startswith('ii'):
                        packages.append(line.split()[1:3])

                # Disconnect from server
                ssh.close()

                if len(packages) > 0:
                    self.packages[serverid]=packages
            except:
                self.LOGGER.error("Server '%s' is unavailale." % serverid)
                self.unavailable_servers.append(serverid)

    @auto_update(15)
    def update_users(self):
        # Get the store
        store=Store(self.db)

        for serverid in self.get_config_path(self.config,"servers"):
            for user in store.find(Account, server=unicode(serverid), enabled=True):
                if (datetime.datetime.now() - user.created).seconds >= int(self.get_config_path(self.config,"servers",serverid,"timelimit")):
                    self.delete_user(user.server,user.username)
        return True

    @auto_update(60)
    def update_servers(self):
        for serverid in self.unavailable_servers:
            try:
                # Connect to the server
                ssh = self.get_ssh(serverid)

                # Test the connection
                ssh.exec_command("echo OK")

                # Disconnect from server
                ssh.close()

                # It worked, so we can remove the server from the blacklist
                self.unavailable_servers.remove(serverid)
                self.LOGGER.info("Server '%s' is back online." % serverid)
            except:
                pass

    def list_users(self,serverid,enabled = True):
        # Get the store
        store=Store(self.db)

        if enabled:
            users=store.find(Account, server=unicode(serverid), enabled=True)
        else:
            users=store.find(Account, server=unicode(serverid))
        userlist=[]
        for user in users:
            userlist.append({'username':str(user.username),'fullname':str(user.fullname),'source':str(user.source),'session':str(user.session),'created':str(user.created)})
        return userlist

    def list_servers(self):
        # Get the store
        store=Store(self.db)

        servers = {}
        for server in self.get_config_path(self.config,"servers"):
            if server in self.unavailable_servers:
                continue

            server_dict={}
            server_dict['title']=self.get_config_path(self.config,"servers",server,"title")
            server_dict['description']=self.get_config_path(self.config,"servers",server,"description")
            server_dict['users']=store.find(Account, server=unicode(server), enabled=True).count()
            server_dict['userlimit']=int(self.get_config_path(self.config,"servers",server,"userlimit"))
            server_dict['timelimit']=int(self.get_config_path(self.config,"servers",server,"timelimit"))
            servers[server]=server_dict

        return servers

    def list_locales(self,serverid):
        if serverid not in self.locales:
            return []

        return self.locales[serverid]

    def list_packages(self,serverid):
        if serverid not in self.packages:
            return []

        return self.packages[serverid]

    def list_everything(self):
        everything={}
        servers=self.list_servers()
        for serverid in servers:
            server=servers[serverid]
            server['locales']=self.list_locales(serverid)
            server['packages']=self.list_packages(serverid)
            everything[serverid]=server

        return everything

    def json_handler(self,query,client):
        reply={}

        # Check if we at least have a function name
        if 'action' not in query:
            reply['status']="error"
            reply['message']=-1
            return reply

        # Check if function is exported over JSON
        if query['action'] not in self.json_functions():
            reply['status']="error"
            reply['message']=-3
            return reply

        # Special case for create_user
        if query['action'] == 'create_user':
            # Make sure we have all parameters
            for param in ('serverid','username','fullname','password','session'):
                if param not in query:
                    reply['status']="error"
                    reply['message']=-2
                    return reply

            # Create the user
            retval=self.create_user(query['serverid'],query['username'],query['fullname'],query['password'],"json:%s" % client[0],query['session'])

            # Check if it worked or if we have an error
            if type(retval) == type([]):
                reply['status']='ok'
                reply['message']=retval
            else:
                reply['status']='error'
                reply['message']=retval
        else:
            # Standard function without parameters, just return output
            reply['status']='ok'
            reply['message']=getattr(self,query['action'])()

        return reply

    def rpc_functions(self):
        return [
            'create_user',
            'delete_user',
            'list_everything',
            'list_locales',
            'list_packages',
            'list_servers',
            'list_users',
        ]

    def json_functions(self):
        return [
            'create_user',
            'list_everything',
            'list_locales',
            'list_packages',
            'list_servers',
        ]
